@@ -2,7 +2,7 @@ from abc import ABC, abstractmethod
 from dataclasses import dataclass
 from typing import Dict, List, Optional, Type
 
-from . import NetQASMInstruction, core, nv, vanilla
+from . import NetQASMInstruction, core, nv, trapped_ion, vanilla
 
 
 @dataclass
@@ -31,7 +31,11 @@ CORE_INSTRUCTIONS: List[Type[NetQASMInstruction]] = [
     core.SubInstruction,
     core.AddmInstruction,
     core.SubmInstruction,
+    core.MulInstruction,
+    core.DivInstruction,
+    core.RemInstruction,
     core.MeasInstruction,
+    core.MeasBasisInstruction,
     core.CreateEPRInstruction,
     core.RecvEPRInstruction,
     core.WaitAllInstruction,
@@ -111,3 +115,29 @@ class NVFlavour(Flavour):
 
     def __init__(self):
         super().__init__(self.instrs)
+
+
+class REIDSFlavour(Flavour):
+    @property
+    def instrs(self):
+        return []
+
+    def __init__(self):
+        super().__init__(self.instrs)
+
+
+class TrappedIonFlavour(Flavour):
+    @property
+    def instrs(self):
+        return [
+            trapped_ion.RotZInstruction,
+            trapped_ion.AllQubitsRotXInstruction,
+            trapped_ion.AllQubitsRotYInstruction,
+            trapped_ion.AllQubitsRotZInstruction,
+            trapped_ion.AllQubitsInitInstruction,
+            trapped_ion.AllQubitsMeasInstruction,
+            trapped_ion.BichromaticInstruction,
+        ]
+
+    def __init__(self):
+        super().__init__(self.instrs)
