@@ -12,44 +12,6 @@ from . import core
 # Explicit instruction types in the NV flavour.
 
 
-@dataclass
-class GateXInstruction(core.SingleQubitInstruction):
-    id: int = 20
-    mnemonic: str = "x"
-
-    def to_matrix(self) -> np.ndarray:
-        return np.array([[0, 1], [1, 0]])
-
-
-@dataclass
-class GateYInstruction(core.SingleQubitInstruction):
-    id: int = 21
-    mnemonic: str = "y"
-
-    def to_matrix(self) -> np.ndarray:
-        return np.array([[0, -1j], [1j, 0]])
-
-
-@dataclass
-class GateZInstruction(core.SingleQubitInstruction):
-    id: int = 22
-    mnemonic: str = "z"
-
-    def to_matrix(self) -> np.ndarray:
-        return np.array([[1, 0], [0, -1]])
-
-
-@dataclass
-class GateHInstruction(core.SingleQubitInstruction):
-    id: int = 23
-    mnemonic: str = "h"
-
-    def to_matrix(self) -> np.ndarray:
-        X = GateXInstruction().to_matrix()
-        Z = GateZInstruction().to_matrix()
-        return (X + Z) / np.sqrt(2)
-
-
 @dataclass
 class RotXInstruction(core.RotationInstruction):
     id: int = 27
@@ -57,7 +19,7 @@ class RotXInstruction(core.RotationInstruction):
 
     def to_matrix(self) -> np.ndarray:
         axis = [1, 0, 0]
-        angle = self.angle_num.value * np.pi / 2 ** self.angle_denom.value
+        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
         return get_rotation_matrix(axis, angle)
 
 
@@ -68,7 +30,7 @@ class RotYInstruction(core.RotationInstruction):
 
     def to_matrix(self) -> np.ndarray:
         axis = [0, 1, 0]
-        angle = self.angle_num.value * np.pi / 2 ** self.angle_denom.value
+        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
         return get_rotation_matrix(axis, angle)
 
 
@@ -79,7 +41,7 @@ class RotZInstruction(core.RotationInstruction):
 
     def to_matrix(self) -> np.ndarray:
         axis = [0, 0, 1]
-        angle = self.angle_num.value * np.pi / 2 ** self.angle_denom.value
+        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
         return get_rotation_matrix(axis, angle)
 
 
@@ -90,12 +52,12 @@ class ControlledRotXInstruction(core.ControlledRotationInstruction):
 
     def to_matrix(self) -> np.ndarray:
         axis = [1, 0, 0]
-        angle = self.angle_num.value * np.pi / 2 ** self.angle_denom.value
+        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
         return get_controlled_rotation_matrix(axis, angle)
 
     def to_matrix_target_only(self) -> np.ndarray:
         axis = [1, 0, 0]
-        angle = self.angle_num.value * np.pi / 2 ** self.angle_denom.value
+        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
         return get_rotation_matrix(axis, angle)
 
 
@@ -106,10 +68,10 @@ class ControlledRotYInstruction(core.ControlledRotationInstruction):
 
     def to_matrix(self) -> np.ndarray:
         axis = [1, 0, 0]
-        angle = self.angle_num.value * np.pi / 2 ** self.angle_denom.value
+        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
         return get_controlled_rotation_matrix(axis, angle)
 
     def to_matrix_target_only(self) -> np.ndarray:
         axis = [1, 0, 0]
-        angle = self.angle_num.value * np.pi / 2 ** self.angle_denom.value
+        angle = self.angle_num.value * np.pi / 2**self.angle_denom.value
         return get_rotation_matrix(axis, angle)
