@@ -28,6 +28,7 @@ class EntRequestParams:
     time_unit: TimeUnit = TimeUnit.MICRO_SECONDS
     max_time: int = 0
     expect_phi_plus: bool = True
+    expect_psi_plus: bool = False
     min_fidelity_all_at_end: Optional[int] = None
     max_tries: Optional[int] = None
     random_basis_local: Optional[RandomBasis] = None
@@ -166,7 +167,7 @@ def deserialize_epr_measure_results(
                 ),
                 measurement_basis_local=request.rotations_local,
                 measurement_basis_remote=request.rotations_remote,
-                post_process=(request.expect_phi_plus and role == EPRRole.CREATE),
+                post_process=(request.expect_phi_plus and role == EPRRole.RECV),
                 remote_node_id=array.get_future_index(
                     base + SER_RESPONSE_MEASURE_IDX_REMOTE_NODE_ID
                 ),
