@@ -522,6 +522,7 @@ class BaseNetQASMConnection(abc.ABC):
             return None
 
         subroutine = self._builder.subrt_compile_subroutine(protosubroutine)
+        self._builder._reset()
 
         return subroutine
 
@@ -556,7 +557,7 @@ class BaseNetQASMConnection(abc.ABC):
         block: bool = True,
         callback: Optional[Callable] = None,
     ) -> None:
-        self._logger.debug(f"Commiting compiled subroutine:\n{subroutine}")
+        self._logger.debug(f"Committing compiled subroutine:\n{subroutine}")
 
         self._commit_message(
             msg=SubroutineMessage(subroutine=subroutine),
