@@ -12,6 +12,7 @@ from netqasm.lang.subroutine import Subroutine
 from netqasm.lang.version import NETQASM_VERSION
 from netqasm.logging.glob import set_log_level
 from netqasm.qlink_compat import EPRType, TimeUnit
+from netqasm.runtime.settings import set_is_using_hardware
 from netqasm.sdk.connection import DebugConnection
 from netqasm.sdk.epr_socket import EPRSocket
 from netqasm.sdk.qubit import Qubit
@@ -157,13 +158,13 @@ def test_rotations():
     print(expected)
 
 
-def test_epr():
+def test_epr_k_create():
 
     set_log_level(logging.DEBUG)
 
     epr_socket = EPRSocket(remote_app_name="Bob")
     with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
-        q1 = epr_socket.create()[0]
+        q1 = epr_socket.create_keep()[0]
         q1.H()
 
     # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
@@ -175,57 +176,30 @@ def test_epr():
     expected_text = """
 # NETQASM 0.0
 # APPID 0
-set R5 10
-array R5 @0
-set R5 1
-array R5 @1
-set R5 0
-set R6 0
-store R5 @1[R6]
-set R5 20
-array R5 @2
-set R5 0
-set R6 0
-store R5 @2[R6]
-set R5 1
-set R6 1
-store R5 @2[R6]
-set R5 1
-set R6 0
-set R7 1
-set R8 2
-set R9 0
-create_epr R5 R6 R7 R8 R9
-set R5 0
-set R6 10
-wait_all @0[R5:R6]
-set R2 0
-set R5 1
-beq R2 R5 51
-load R0 @1[R2]
-set R3 9
-set R4 0
-beq R4 R2 36
-set R5 10
-add R3 R3 R5
-set R5 1
-add R4 R4 R5
-jmp 30
-load R1 @0[R3]
+set R0 10
+array R0 @0
+set R0 1
+array R0 @1
 set R0 0
-set R5 3
-bne R1 R5 41
-rot_z R0 16 4
-set R5 1
-bne R1 R5 44
-rot_x R0 16 4
-set R5 2
-bne R1 R5 48
-rot_x R0 16 4
-rot_z R0 16 4
-set R5 1
-add R2 R2 R5
-jmp 25
+set R1 0
+store R0 @1[R1]
+set R0 20
+array R0 @2
+set R0 0
+set R1 0
+store R0 @2[R1]
+set R0 1
+set R1 1
+store R0 @2[R1]
+set R0 1
+set R1 0
+set R2 1
+set R3 2
+set R4 0
+create_epr R0 R1 R2 R3 R4
+set R0 0
+set R1 10
+wait_all @0[R0:R1]
 set Q0 0
 h Q0
 ret_arr @0
@@ -245,7 +219,171 @@ ret_arr @2
     print(expected)
 
 
-def test_two_epr():
+def test_epr_k_recv():
+
+    set_log_level(logging.DEBUG)
+
+    epr_socket = EPRSocket(remote_app_name="Bob")
+    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
+        q1 = epr_socket.recv_keep()[0]
+        q1.H()
+
+    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
+    assert len(alice.storage) == 5
+    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
+    subroutine = deserialize_subroutine(raw_subroutine)
+    print(subroutine)
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
+set R5 10
+array R5 @0
+set R5 1
+array R5 @1
+set R5 0
+set R6 0
+store R5 @1[R6]
+set R5 1
+set R6 0
+set R7 1
+set R8 0
+recv_epr R5 R6 R7 R8
+set R5 0
+set R6 10
+wait_all @0[R5:R6]
+set R2 0
+set R5 1
+beq R2 R5 42
+load R0 @1[R2]
+set R3 9
+set R4 0
+beq R4 R2 27
+set R5 10
+add R3 R3 R5
+set R5 1
+add R4 R4 R5
+jmp 21
+load R1 @0[R3]
+set R0 0
+set R5 3
+bne R1 R5 32
+rot_z R0 16 4
+set R5 1
+bne R1 R5 35
+rot_x R0 16 4
+set R5 2
+bne R1 R5 39
+rot_x R0 16 4
+rot_z R0 16 4
+set R5 1
+add R2 R2 R5
+jmp 16
+set Q0 0
+h Q0
+ret_arr @0
+ret_arr @1
+ret_arr @2
+"""
+
+    expected = parse_text_subroutine(expected_text)
+
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
+        print()
+        assert instr == expected_instr
+    print(subroutine)
+    print(expected)
+
+
+def test_epr_k_recv_hardware():
+    set_is_using_hardware(True)
+
+    set_log_level(logging.DEBUG)
+
+    epr_socket = EPRSocket(remote_app_name="Bob")
+    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
+        q1 = epr_socket.recv_keep()[0]
+        q1.H()
+
+    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
+    assert len(alice.storage) == 5
+    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
+    subroutine = deserialize_subroutine(raw_subroutine)
+    print(subroutine)
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
+set R5 10
+array R5 @0
+set R5 1
+array R5 @1
+set R5 0
+set R6 0
+store R5 @1[R6]
+set R5 1
+set R6 0
+set R7 1
+set R8 0
+recv_epr R5 R6 R7 R8
+set R5 0
+set R6 10
+wait_all @0[R5:R6]
+set R2 0
+set R5 1
+beq R2 R5 46
+load R0 @1[R2]
+set R3 9
+set R4 0
+beq R4 R2 27
+set R5 10
+add R3 R3 R5
+set R5 1
+add R4 R4 R5
+jmp 21
+load R1 @0[R3]
+set R0 0
+set R5 3
+bne R1 R5 34
+rot_y R0 8 4
+rot_x R0 16 4
+rot_y R0 24 4
+set R5 1
+bne R1 R5 37
+rot_x R0 16 4
+set R5 2
+bne R1 R5 43
+rot_x R0 16 4
+rot_y R0 8 4
+rot_x R0 16 4
+rot_y R0 24 4
+set R5 1
+add R2 R2 R5
+jmp 16
+set Q0 0
+h Q0
+ret_arr @0
+ret_arr @1
+"""
+
+    expected = parse_text_subroutine(expected_text)
+
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
+        print()
+        assert instr == expected_instr
+    print(subroutine)
+    print(expected)
+
+    set_is_using_hardware(False)
+
+
+def test_two_epr_k_create():
 
     set_log_level(logging.DEBUG)
 
@@ -264,6 +402,72 @@ def test_two_epr():
     expected_text = """
 # NETQASM 0.0
 # APPID 0
+set R0 20
+array R0 @0
+set R0 2
+array R0 @1
+set R0 0
+set R1 0
+store R0 @1[R1]
+set R0 1
+set R1 1
+store R0 @1[R1]
+set R0 20
+array R0 @2
+set R0 0
+set R1 0
+store R0 @2[R1]
+set R0 2
+set R1 1
+store R0 @2[R1]
+set R0 1
+set R1 0
+set R2 1
+set R3 2
+set R4 0
+create_epr R0 R1 R2 R3 R4
+set R0 0
+set R1 20
+wait_all @0[R0:R1]
+set Q0 0
+h Q0
+set Q0 1
+h Q0
+ret_arr @0
+ret_arr @1
+ret_arr @2
+    """
+
+    expected = parse_text_subroutine(expected_text)
+
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
+        assert instr == expected_instr
+    print(subroutine)
+    print(expected)
+
+
+def test_two_epr_k_recv():
+
+    set_log_level(logging.DEBUG)
+
+    epr_socket = EPRSocket(remote_app_name="Bob")
+    with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
+        qubits = epr_socket.recv(number=2)
+        qubits[0].H()
+        qubits[1].H()
+
+    # 5 messages: init, open_epr_socket, subroutine, stop app and stop backend
+    assert len(alice.storage) == 5
+    raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
+    subroutine = deserialize_subroutine(raw_subroutine)
+    print(subroutine)
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
 set R5 20
 array R5 @0
 set R5 2
@@ -274,50 +478,41 @@ store R5 @1[R6]
 set R5 1
 set R6 1
 store R5 @1[R6]
-set R5 20
-array R5 @2
-set R5 0
-set R6 0
-store R5 @2[R6]
-set R5 2
-set R6 1
-store R5 @2[R6]
 set R5 1
 set R6 0
 set R7 1
-set R8 2
-set R9 0
-create_epr R5 R6 R7 R8 R9
+set R8 0
+recv_epr R5 R6 R7 R8
 set R5 0
 set R6 20
 wait_all @0[R5:R6]
 set R2 0
 set R5 2
-beq R2 R5 54
+beq R2 R5 45
 load R0 @1[R2]
 set R3 9
 set R4 0
-beq R4 R2 39
+beq R4 R2 30
 set R5 10
 add R3 R3 R5
 set R5 1
 add R4 R4 R5
-jmp 33
+jmp 24
 load R1 @0[R3]
 set R0 0
 set R5 3
-bne R1 R5 44
+bne R1 R5 35
 rot_z R0 16 4
 set R5 1
-bne R1 R5 47
+bne R1 R5 38
 rot_x R0 16 4
 set R5 2
-bne R1 R5 51
+bne R1 R5 42
 rot_x R0 16 4
 rot_z R0 16 4
 set R5 1
 add R2 R2 R5
-jmp 28
+jmp 19
 set Q0 0
 h Q0
 set Q0 1
@@ -528,155 +723,44 @@ def test_epr_r_create():
     assert len(alice.storage) == 5
     raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
-    expected = Subroutine(
-        netqasm_version=NETQASM_VERSION,
-        app_id=0,
-        instructions=[
-            # Arg array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(OK_FIELDS),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 1),
-                address=Address(0),
-            ),
-            # ent info array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(CREATE_FIELDS),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 1),
-                address=Address(1),
-            ),
-            # tp arg
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(2),  # EPRType.R
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 2),
-                imm=Immediate(0),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 1),
-                entry=ArrayEntry(1, index=Register(RegisterName.R, 2)),
-            ),
-            # num pairs arg
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 2),
-                imm=Immediate(1),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 1),
-                entry=ArrayEntry(1, index=Register(RegisterName.R, 2)),
-            ),
-            # create cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 2),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 3),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 4),
-                imm=Immediate(0),
-            ),
-            instructions.core.CreateEPRInstruction(
-                reg0=Register(RegisterName.R, 1),
-                reg1=Register(RegisterName.R, 2),
-                reg2=Register(RegisterName.C, 0),
-                reg3=Register(RegisterName.R, 3),
-                reg4=Register(RegisterName.R, 4),
-            ),
-            # wait cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 2),
-                imm=Immediate(OK_FIELDS),
-            ),
-            instructions.core.WaitAllInstruction(
-                slice=ArraySlice(
-                    address=Address(0),
-                    start=Register(RegisterName.R, 1),
-                    stop=Register(RegisterName.R, 2),
-                ),
-            ),
-            # if statement
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(2),
-            ),
-            instructions.core.LoadInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(
-                    address=Address(0),
-                    index=Register(RegisterName.R, 1),
-                ),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.BneInstruction(
-                reg0=Register(RegisterName.R, 0),
-                reg1=Register(RegisterName.R, 1),
-                imm=Immediate(28),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(2),
-            ),
-            instructions.core.LoadInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(
-                    address=Address(0),
-                    index=Register(RegisterName.R, 1),
-                ),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(1),
-            ),
-            instructions.core.AddInstruction(
-                reg0=Register(RegisterName.R, 0),
-                reg1=Register(RegisterName.R, 0),
-                reg2=Register(RegisterName.R, 1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(2),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(
-                    address=Address(0),
-                    index=Register(RegisterName.R, 1),
-                ),
-            ),
-            # return cmds
-            instructions.core.RetArrInstruction(
-                address=Address(0),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(1),
-            ),
-        ],
-    )
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
+set R1 10
+array R1 @0
+set R1 20
+array R1 @1
+set R1 2
+set R2 0
+store R1 @1[R2]
+set R1 1
+set R2 1
+store R1 @1[R2]
+set R1 1
+set R2 0
+set R3 1
+set R4 0
+create_epr R1 R2 C0 R3 R4
+set R1 0
+set R2 10
+wait_all @0[R1:R2]
+set R1 2
+load R0 @0[R1]
+set R1 0
+bne R0 R1 28
+set R1 2
+load R0 @0[R1]
+set R1 1
+add R0 R0 R1
+set R1 2
+store R0 @0[R1]
+ret_arr @0
+ret_arr @1
+    """
+
+    expected = parse_text_subroutine(expected_text)
+
     for i, instr in enumerate(subroutine.instructions):
         print(repr(instr))
         expected_instr = expected.instructions[i]
@@ -700,87 +784,58 @@ def test_epr_r_receive():
     raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
     print(subroutine)
-    expected = Subroutine(
-        netqasm_version=NETQASM_VERSION,
-        app_id=0,
-        instructions=[
-            # Arg array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(OK_FIELDS),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(0),
-            ),
-            # Qubit address array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(1),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(1, index=Register(RegisterName.R, 1)),
-            ),
-            # create cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 2),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 3),
-                imm=Immediate(0),
-            ),
-            instructions.core.RecvEPRInstruction(
-                reg0=Register(RegisterName.R, 0),
-                reg1=Register(RegisterName.R, 1),
-                reg2=Register(RegisterName.R, 2),
-                reg3=Register(RegisterName.R, 3),
-            ),
-            # wait cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(OK_FIELDS),
-            ),
-            instructions.core.WaitAllInstruction(
-                slice=ArraySlice(
-                    address=Address(0),
-                    start=Register(RegisterName.R, 0),
-                    stop=Register(RegisterName.R, 1),
-                ),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(0),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(1),
-            ),
-        ],
-    )
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
+set R5 10
+array R5 @0
+set R5 1
+array R5 @1
+set R5 0
+set R6 0
+store R5 @1[R6]
+set R5 1
+set R6 0
+set R7 1
+set R8 0
+recv_epr R5 R6 R7 R8
+set R5 0
+set R6 10
+wait_all @0[R5:R6]
+set R2 0
+set R5 1
+beq R2 R5 42
+load R0 @1[R2]
+set R3 9
+set R4 0
+beq R4 R2 27
+set R5 10
+add R3 R3 R5
+set R5 1
+add R4 R4 R5
+jmp 21
+load R1 @0[R3]
+set R0 0
+set R5 3
+bne R1 R5 32
+rot_z R0 16 4
+set R5 1
+bne R1 R5 35
+rot_x R0 16 4
+set R5 2
+bne R1 R5 39
+rot_x R0 16 4
+rot_z R0 16 4
+set R5 1
+add R2 R2 R5
+jmp 16
+ret_arr @0
+ret_arr @1
+    """
+
+    expected = parse_text_subroutine(expected_text)
+
     for i, instr in enumerate(subroutine.instructions):
         print(f"actual command: {repr(instr)}")
         expected_instr = expected.instructions[i]
@@ -809,63 +864,36 @@ def test_epr_max_time():
     expected_text = """
 # NETQASM 0.0
 # APPID 0
-set R5 10
-array R5 @0
-set R5 1
-array R5 @1
-set R5 0
-set R6 0
-store R5 @1[R6]
-set R5 20
-array R5 @2
-set R5 0
-set R6 0
-store R5 @2[R6]
-set R5 1
-set R6 1
-store R5 @2[R6]
-set R5 1
-set R6 5
-store R5 @2[R6]
-set R5 25
-set R6 6
-store R5 @2[R6]
-set R5 1
-set R6 0
-set R7 1
-set R8 2
-set R9 0
-create_epr R5 R6 R7 R8 R9
-set R5 0
-set R6 10
-wait_all @0[R5:R6]
-set R2 0
-set R5 1
-beq R2 R5 57
-load R0 @1[R2]
-set R3 9
-set R4 0
-beq R4 R2 42
-set R5 10
-add R3 R3 R5
-set R5 1
-add R4 R4 R5
-jmp 36
-load R1 @0[R3]
+set R0 10
+array R0 @0
+set R0 1
+array R0 @1
 set R0 0
-set R5 3
-bne R1 R5 47
-rot_z R0 16 4
-set R5 1
-bne R1 R5 50
-rot_x R0 16 4
-set R5 2
-bne R1 R5 54
-rot_x R0 16 4
-rot_z R0 16 4
-set R5 1
-add R2 R2 R5
-jmp 31
+set R1 0
+store R0 @1[R1]
+set R0 20
+array R0 @2
+set R0 0
+set R1 0
+store R0 @2[R1]
+set R0 1
+set R1 1
+store R0 @2[R1]
+set R0 1
+set R1 5
+store R0 @2[R1]
+set R0 25
+set R1 6
+store R0 @2[R1]
+set R0 1
+set R1 0
+set R2 1
+set R3 2
+set R4 0
+create_epr R0 R1 R2 R3 R4
+set R0 0
+set R1 10
+wait_all @0[R0:R1]
 set Q0 0
 h Q0
 ret_arr @0
@@ -887,8 +915,10 @@ ret_arr @2
 if __name__ == "__main__":
     test_simple()
     test_rotations()
-    test_epr()
-    test_two_epr()
+    test_epr_k_create()
+    test_epr_k_recv()
+    test_two_epr_k_create()
+    test_two_epr_k_recv()
     test_epr_m()
     test_epr_r_create()
     test_epr_r_receive()
